/* This program reads a coded bibfile of a conference and:

   (a) computes the clusters at the end of each edition
   (b) determines if a paper creates a new cluster, extends an existing one
   (c) connects internally authors that already belong to one cluster
   (d) merges two or more existing clusters

 It outputs what does each paper do and a summary at the end of the edition
*/

#include "stdio.h"
#include "ctype.h"
#include "string.h"


#define MAXLINE 400   /* Number of characters on a line */
#define MAXTITLESIZE 500 /* Max number of characters on a title */
#define ACODESZ 8 /* Author code size */
#define MAXCOAUTHORS 30 /* Max number of coauthors per article */
#define MAXKEY 12 /* Max length of key */
#define FILESTR 20 /* String length of file names */

#define MAXAUTHORS 10000 /* Max number of authors in database*/
#define MAXEDITIONS 16 /* Maximum number of editions of the conference */

#define NUL '\0'
#define QTE '"' 

typedef int boolean;
#define TRUE 1
#define FALSE 0

/* Types of papers */

#define S 0 /* Start (auxiliary) */
#define N 1 /* New: all authors are newcomers */
#define I 2 /* Internal: all authors belonged to one cluster */
#define E 3 /* Extension: all authors are new ore bleonged to one cluster */
#define M 4 /* Merge: merges two or more existing clusters (plus perhaps newcomers */

char inpline[MAXLINE];
int cursor;

int lineno;

/*char authorline[MAXLINE];*/

char prefix[MAXKEY];

int art_auth[MAXCOAUTHORS]; /*Author list for a publication*/

int maxauth; /* Author with largest number*/
															
															/* Cluster matrix: every author has the cluster number she belongs to 
   each edition. Editions start at 1, if years not consecutive, not 
   the same as curr_Year-First_year*/

int ClMtx[MAXEDITIONS][MAXAUTHORS];
int CurrEd;

/* Merge cluster array. number of clusters that merge with a paper */
int mergeCl[MAXCOAUTHORS];
int mergeCl_nr;

/* Type of cluster accumulated data */
int TypeMtx[5][MAXEDITIONS];

int nextClnr; /* Next (new) cluster number */

FILE * BIBF; /* the input bib file */
FILE * PLOGF; /* the paper log data outfile */
FILE * CLF;  /* The cluster output file */
FILE * CLSUMF;  /* The cluster summary file */


/* ------------ Functions ------------------- */

int readln(f,s)
FILE * f;
char *s;

/* Reads a line from file f and puts it into string s. Ends it with \0    */
/* If all is OK returns the number of characters in s, if EOF is reached, */
/* it returns 0, otherwise returns -1. Truncates line to MAXLINE          */
/* characters. MAXLINE is set to 80 if it is not defined previously       */

#ifndef MAXLINE
#define MAXLINE 80
#endif

{ int i, ok;
  char *p, ch;
  
  lineno++;
  
  p = s; i = 0;
  while (i < MAXLINE)
   { ok = fscanf(f,"%c",&ch);
     if (ok == 1) /* All is well */
       { if (ch == '\n') /* end of line */
           { *p = NUL;
             return(i+1);
            }
         *p++ = ch;
         i++;
       }
      else if (ok == EOF) 
            { *p = NUL;
              return(0);
             }
            else return(-1); /* ok has a strange value */ 
   }
  *p = NUL;
  return(MAXLINE);
}

/* ------------------ */

boolean isauthorline()

/* Returns TRUE if first non blank characters are 'AUTHOR', FALSE otherwise */

{ int i;
	
  i = 0;
  while(inpline[i] == ' ') i++; /* jump blanks */
  
  /* Just in case, check that we don't run out of characters...*/
  
  if ((inpline[i] == NUL) ||  (inpline[i+1] == NUL) || (inpline[i+2] == NUL) ||
	    (inpline[i+3] == NUL) || (inpline[i+4] == NUL) || (inpline[i+5] == NUL))
    return(FALSE);

  /* Is it? */
  
  if ((tolower(inpline[i]) == 'a') &&  (tolower(inpline[i+1]) == 'u') && 
       (tolower(inpline[i+2]) == 't') && (tolower(inpline[i+3]) == 'h') && 
       (tolower(inpline[i+4]) == 'o') && (tolower(inpline[i+5]) == 'r'))
    return(TRUE);

  /* No, it isn't */
  
  return(FALSE);
}
	

/* ------------------------------------ */

boolean istitleline()

/* Returns TRUE if first non blank characters are 'TITLE', FALSE otherwise */

{ int i;
	
  i = 0;
  while(inpline[i] == ' ') i++; /* jump blanks */
  
  /* Just in case, check that we don't run out of characters...*/
  
  if ((inpline[i] == NUL) ||  (inpline[i+1] == NUL) || (inpline[i+2] == NUL) ||
	    (inpline[i+3] == NUL) || (inpline[i+4] == NUL))
    return(FALSE);

  /* Is it? */
  
  if ((tolower(inpline[i]) == 't') &&  (tolower(inpline[i+1]) == 'i') && 
        (tolower(inpline[i+2]) == 't') && (tolower(inpline[i+3]) == 'l') &&
	    (tolower(inpline[i+4]) == 'e'))
    return(TRUE);

  /* No, it isn't */
  
  return(FALSE);
}
	
/* ------------------------------------ */

boolean isyearline()

/* Returns TRUE if first non blank characters are 'YEAR', FALSE otherwise */

{ int i;
	
  i = 0;
  while(inpline[i] == ' ') i++; /* jump blanks */
  
  /* Just in case, check that we don't run out of characters...*/
  
  if ((inpline[i] == NUL) ||  (inpline[i+1] == NUL) || (inpline[i+2] == NUL) ||
	    (inpline[i+3] == NUL))
    return(FALSE);

  /* Is it? */
  
  if ((tolower(inpline[i]) == 'y') &&  (tolower(inpline[i+1]) == 'e') && 
       (tolower(inpline[i+2]) == 'a') && (tolower(inpline[i+3]) == 'r'))
    return(TRUE);

  /* No, it isn't */
  
  return(FALSE);
}

/* ------------------------------------ */


void get_bibkey(s)
char s[];

/* Gets bib_ref key from input line, and puts it into s. */
/* Line always ends with ,                               */ 

{ int i,j;

  i = 0; j = 0;
  while (inpline[i++] != '{'); /* Go to key */
  while (inpline[i] == ' ') i++;
  
  while ((inpline[i] != ',') && (inpline[i] != ' ')) 
	   s[j++] = inpline[i++];
  s[j] = NUL;
}

/* ------------------------------------ */


int get_year()


/* Gets year  from input line, Already checked it is a YEAR line */

{ int i,j;

  i = 0; j = 0;
 
  /* go to first digit */
  while (!isdigit(inpline[i])) i++;
  j = inpline[i++] -'0';
  j = j*10 + (inpline[i++] -'0');
  j = j*10 + (inpline[i++] -'0');
  j = j*10 + (inpline[i] -'0');
  return(j);
}




/* ------------------------------------ */

int get_author_number()

/* Gets the number of the first author it finds from aline starting at cursor. 
Returns the number (from 0 to 9999) or -1 if there is no more authors*/

{ int n;

  /* Go to beginning of author code */
  while ((inpline[cursor] != prefix[0]) && (inpline[cursor] != NUL)) cursor++;
  
  if (inpline[cursor] == NUL) return(-1); /* No more authors*/
  /* Will not check rest of prefix. Beware*/
   n = (inpline[cursor+3] - '0')*1000 + (inpline[cursor+4] - '0')*100 +
         (inpline[cursor+5] - '0')*10 + (inpline[cursor+6] - '0');
		 
   cursor = cursor + 7; /*Prepared for next*/
   
   /* Update max author if necessary */
   if (n > maxauth) maxauth = n;
   
   return(n);
 }
   
   

/* ------------------------------------ */

int get_all_authors()

/* gets all autors in inpline[] and puts their numbers (from 1 to 9999) in
   art_auth[]. The rest of the entries are -1. Returns number of authors for the 
   publication. */
   
{ int i;
  int AuNr, n_of_auth; /* Author number and Number of authors in the publication */
  
  for (i = 0; i < MAXCOAUTHORS; i++) art_auth[i] = -1; /* Wipe out just in case */
  
  /* Go to first author*/
  
  cursor = 0;
  
  while (inpline[cursor++] != QTE);
  n_of_auth = 0;
  
  /* Get authors */
  AuNr = get_author_number();
  
  while (AuNr != -1)
    { art_auth[n_of_auth++] = AuNr;
	  AuNr = get_author_number();
	}
  return(n_of_auth);
}
  
	  
/* ------------------------------------ */

int MaxClNr(n)
int n;

/* Obtains the maximum cluster number of the n authors in the maxauth 
 * array */

{ int mx, i;

  mx = 0;
  
  for (i = 0; i < n; i++) 
	  if (ClMtx[CurrEd][art_auth[i]] > mx) 
		  mx = ClMtx[CurrEd][art_auth[i]];
  return(mx);
}



/*___________*/


void updateCurrCl(na)
int na; /* Number of authors in current paper */
  
/* From paper data obtained in get_all_authors
   updates current cluster list with data */

{ int i, j;
  int mc; /* Max cluster number */
  int clnrPA; /* Cluster number of present author */

  mc = MaxClNr();
  
  if (mc == 0) /* If cluster max = 0, new cluster number */
	  { mc = nextClnr++;
	    for (i = 0; i < na; i++) ClMtx[CurrEd][art_auth[i]] = mc;
		return;
	  }
  /* At least one is of an existing cluster */
  /* For each coauthor */
  for (i = 0; i < na; i++)
	  { clnrPA = ClMtx[CurrEd][art_auth[i]];
	    /* If new author, set it to mc */
		if (clnrPA == 0) ClMtx[CurrEd][art_auth[i]] = mc;
		  else /* if not new author*/
		  if (clnrPA != mc) /* if of different cluster set all authors of that cluster to new cluster */
		     for (j = 1; j <= maxauth; j++) if (ClMtx[CurrEd][j] == clnrPA) ClMtx[CurrEd][j] = mc;
	  }
}	

/*___________*/

boolean inlist(cl)
int cl;

/* Checks if cluster cl is in the list of merged clusters */

{ int i;
  boolean result;
  
  result = FALSE;

  for (i = 0; i < mergeCl_nr; i++) if (mergeCl[i] == cl) result  = TRUE;

  return(result);
}


/*___________*/

void PaperType(key, year, n_au)
char key[]; /* The paper key */
int year; /* Paper year */
int n_au; /* number of authors of the paper */

/* Finds out the type of the paper. Author list is already in art_auth.
   nextClnr is of a yet unasigned cluster. Prints out the paper 
   information */

{ int i;
  int ext_int; /* cluster of Extension or Internal types */
  int aucl; /* Cluster the author belonged to past edition*/
  int typecl; /* Cluster tye */
  
  /* Initialize */
  typecl = S; ext_int = 0; mergeCl_nr = 0;
  
  /* For each author of the paper */
  for (i = 0; i < n_au; i++)
	  { aucl = ClMtx[CurrEd-1][art_auth[i]];
	    
	    switch (typecl)
	      { case S: if (aucl == 0) typecl = N; 
		              else 
					    { typecl = I;
						  ext_int = aucl;
					    }
					 break;
			case N: if (aucl != 0)
			          { typecl = E;
					    ext_int = aucl;
			          }
			          
					 break;
			case I: if (aucl == 0) typecl = E; /* Intern becomes Extension */
			         else
					   if (aucl != ext_int) /* It's a merge of ext_int and aucl */
					     { typecl = M;
						   mergeCl[0] = ext_int;
						   mergeCl[1] = aucl;
						   mergeCl_nr = 2;
					     }
					 /* If aucl != 0 and != ext_int no change */
					 break;
			case E: if (aucl != ext_int) /* It's a merge of ext_int and aucl */
					     { typecl = M;
						   mergeCl[0] = ext_int;
						   mergeCl[1] = aucl;
						   mergeCl_nr = 2;
					     }
					 /* If aucl == 0 or == ext_int no change */
					 break;
			case M: if ((aucl != 0) && (!inlist(aucl)))  /* A new custer. Add */
					     mergeCl[mergeCl_nr++] = aucl;
					break;
	      }
	  }
   /* Final type established. Add to data */
   TypeMtx[typecl][CurrEd]++;
   
   /* Print results */
   fprintf(PLOGF, "%s\t%d\t",key,year);
   switch (typecl)
	 { case N: fprintf(PLOGF, "New\n");
			   break;
	   case I: fprintf(PLOGF, "Internal\t%d\n", ext_int);
			   break;
	   case E: fprintf(PLOGF, "Extension\t%d\n", ext_int);
			   break;
	   case M: fprintf(PLOGF, "Merge");
	           for (i = 0; i < mergeCl_nr; i++)  
	              fprintf(PLOGF,"\t%d", mergeCl[i]);
	           fprintf(PLOGF, "\n");
			   break;
     }
}

/* ------------------------------------ */
	

main(int argc, char *argv[])
{ 
  char  key[MAXKEY];
  int CurrYear, year;
  int col, totAuNr;
  
  char bibfn[FILESTR], logfn[FILESTR], clusterfn[FILESTR], clsumfn[FILESTR];
  
  boolean eof; /* End of file */
  /* 
   * boolean gkey, gaut, gtit, gyear;
   */

  int auxl,i,j;
  int ed, au;

  /* Initialize */  
  lineno = 0;
  maxauth = 0;
  
  for (ed = 0; ed < MAXEDITIONS; ed++)
	  for (au = 0; au < MAXAUTHORS; au++)
		  ClMtx[ed][au] = 0;
  CurrEd = 1;


  /* Type of cluster accumulated data */
  for ( i = 0; i < 5; i++)
	  for (ed = 0; ed < MAXEDITIONS; ed++)
		  TypeMtx[i][ed] = 0;

  nextClnr = 1; 

  /* Check input line and initialize*/
  if (argc != 2) 
    {printf("usage: %s prefix\n", argv[0] );
	exit(0);
  }
  
  strcpy(prefix,argv[1]); /* Careful.. using same  prefix for authors and filenames */
  
  /* Form file names */
  strcpy(bibfn,argv[1]);
  strcat(bibfn, "Code.bib");
  
  strcpy(logfn,argv[1]);
  strcat(logfn, "PapType.log");

  strcpy(clusterfn,argv[1]);
  strcat(clusterfn, "cluster.csv");

  strcpy(clsumfn,argv[1]);
  strcat(clsumfn, "cldata.csv");

  /* Open files */
  BIBF = fopen(bibfn,"r");
  PLOGF = fopen(logfn,"w");   
  CLF = fopen(clusterfn,"w");
  CLSUMF = fopen(clsumfn,"w"); 
  
  
  /* Test 2 */
  /* Get first entry*/
  auxl = readln(BIBF,inpline); 
  while (inpline[0] != '@') auxl = readln(BIBF,inpline);
  get_bibkey(key);
  /* title and author might be in either order. No need for title */
  auxl = readln(BIBF,inpline);
  if (isauthorline()) totAuNr = get_all_authors();
  auxl = readln(BIBF,inpline);
  if (isauthorline()) totAuNr = get_all_authors();
    /* Whatever order we have title and author */
  auxl = readln(BIBF,inpline);
  year = get_year();
  /* Got forst entry */
  CurrYear = year;
  
  eof = FALSE;
    
  while (!eof)
    { while ((CurrYear == year) && !eof)
       { updateCurrCl(totAuNr);
	
          /* Get next entry. Beware of End of File*/
	      while ((inpline[0] != '@') && (auxl != 0)) auxl = readln(BIBF,inpline);
		  if (auxl == 0) eof = TRUE;
		    else { /* Not end of file, read next entry */
              get_bibkey(key);

              /* title and author might be in either order. No need for title */
              auxl = readln(BIBF,inpline);
              if (isauthorline()) totAuNr = get_all_authors();
              auxl = readln(BIBF,inpline);
              if (isauthorline()) totAuNr = get_all_authors();
              /* Whatever order we have title and author */
              auxl = readln(BIBF,inpline);
              year = get_year();
              }
		
       }
       if (! eof) /* End of year */
	     { 
		   /* Initialize clusters of next edition */
		   for (i = 1; i <= maxauth; i++) ClMtx[CurrEd+1][i] = ClMtx[CurrEd][i];
		   
		   /* Update edition and year */
		   CurrEd++;
		   CurrYear = year;
	     }
    }
    		
  /* Ready to print cluster information. First the header */
  fprintf(CLF,"Edition 1");
  for (ed = 2; ed <= CurrEd; ed++) fprintf(CLF,"\tEdition %d",ed);
  fprintf(CLF,"\n");
  
  /* Now all the clusters of each author per  edition */
  for (au = 1; au <= maxauth; au++)
    { fprintf(CLF,"%d",ClMtx[1][au]);
	  for (ed = 2; ed <= CurrEd; ed++) fprintf(CLF,"\t%d",ClMtx[ed][au]);
	  fprintf(CLF,"\n");
    }
    
  
  fclose(BIBF);
  fclose(PLOGF);
  fclose(CLF);
  fclose(CLSUMF);

}  
